#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>


#include "states.h"
#include "log.h"

#if LIBDBI_VERSION >= 900
static dbi_inst dbi_instance;
# define dbi_initialize(dir) dbi_initialize_r(dir, &dbi_instance)
# define dbi_driver_open(name) dbi_driver_open_r(name, dbi_instance)
# define dbi_shutdown() dbi_shutdown_r(dbi_instance)
#endif

static dbi_conn connection = NULL;

static char *nagios_path = NULL;

int
nagios_connect(char *path)
{
	nagios_path = path;
	return 0;
}

static int
submit_service(const proc_result *result, int new_state)
{
	char *cmd;
	int fd;
	if (!nagios_path) {
		log_handler(LOG_ERR, result->handler->name, "Nagios pipe path not configured, cannot submit command");
		return 1;
	}
	if (asprintf(&cmd, "[%u] PROCESS_SERVICE_CHECK_RESULT;%s;%s;%u;%s\n",
			(unsigned int)result->trap->timestamp, result->host, result->service, new_state, result->message) < 0) {
		log_handler(LOG_ERR, result->handler->name, "Error formating nagios command");
		return 1;
	}
	log_handler(LOG_DEBUG, result->handler->name, "Notifying: %s", cmd);
	fd = open(nagios_path, O_WRONLY|O_APPEND);
	if (fd == -1) {
		log_handler(LOG_ERR, result->handler->name, "Nagios pipe couldn't be opened, cannot submit command");
		return 1;
	}
	if (write(fd, cmd, strlen(cmd)) < 0)
		log_handler(LOG_ERR, result->handler->name, "Couldn't write to nagios pipe, command not submitted");
	close(fd);
	return 0;
}

static enum result_state_t
determine_service_state(const char *host, const char *service)
{
	enum result_state_t state = STATE_UNCHANGED;
	char *h, *s;
	dbi_conn_quote_string_copy(connection, host, &h);
	dbi_conn_quote_string_copy(connection, service, &s);
	dbi_result dbr = dbi_conn_queryf(connection,
			"SELECT MAX(state) AS state FROM subservices "
			"WHERE host = %s AND service = %s",
			h, s);
	free(h);
	free(s);
	if (!dbr || !dbi_result_get_numrows(dbr))
		return state;
	dbi_result_next_row(dbr);
	state = dbi_result_get_int(dbr, "state");
	dbi_result_free(dbr);
	return state;
}

static int
set_substate(const proc_result *result)
{
	log_msg(LOG_DEBUG, "Inserting %s;%s;%s", result->host, result->service, result->tag);
	char *host, *service, *tag, *message, *oid, *trapdata;
	dbi_conn_quote_string_copy(connection, result->host, &host);
	dbi_conn_quote_string_copy(connection, result->service, &service);
	dbi_conn_quote_string_copy(connection, result->tag, &tag);
	dbi_conn_quote_string_copy(connection, result->message, &message);
	dbi_conn_quote_string_copy(connection, result->trap->oid, &oid);
	char *tmp = display_fields(result->trap->fields);
	dbi_conn_quote_string_copy(connection, tmp, &trapdata);
	dbi_result dbr = dbi_conn_queryf(connection,
			"INSERT INTO subservices(host, service, tag, state, message, count, trapoid, trap, time) "
			"VALUES(%s, %s, %s, %d, %s, 1, %s, %s, FROM_UNIXTIME(%d)) "
			"ON DUPLICATE KEY UPDATE count = if(state <> %d, 1, count + 1), state = %d, "
			"message = %s, trapoid = %s, trap = %s, time = FROM_UNIXTIME(%d)",
			host, service, tag, result->state, message, oid, trapdata, result->trap->timestamp,
			result->state, result->state,
			message, oid, trapdata, result->trap->timestamp);
	free(tmp);
	free(host);
	free(service);
	free(tag);
	free(message);
	free(oid);
	free(trapdata);
	if (dbr == NULL)
		return -1;
	dbi_result_free(dbr);
	return 0;
}

int
update_state(const proc_result *result)
{
	if (result->state == STATE_UNCHANGED) {
		log_handler(LOG_INFO, result->handler->name, "No state change");
		return 0;
	}
	int old_state = determine_service_state(result->host, result->service);
	if (set_substate(result)) {
		const char *errstr;
		dbi_conn_error(connection, &errstr);
		log_handler(LOG_ERR, result->handler->name, "Failed to insert subservice: %s", errstr);
	}
	int new_state = determine_service_state(result->host, result->service);
	if (old_state != new_state || result->notify == NOTIFY_ALWAYS) {
		log_handler(LOG_DEBUG, result->handler->name, "Submitting result to nagios");
		submit_service(result, new_state);
	}
	return 0;
}

dbi_conn
db_connect(char *host, char *username, char *password, char *dbname, int port)
{
	int r;
	if (connection != NULL) {
		log_msg(LOG_ERR, "Already connected. This shouldn't happen.");
		return NULL;
	}
	dbi_initialize(NULL);
	dbi_driver driver = dbi_driver_open("mysql");
	if (!driver) {
		log_msg(LOG_ERR, "Couldn't find mysql database driver");
		return NULL;
	}
	dbi_conn conn = dbi_conn_open(driver);
	r = dbi_conn_set_option(conn, "host", host);
	if (r) {
		log_msg(LOG_ERR, "Error setting host");
		return NULL;
	}
	r = dbi_conn_set_option(conn, "username", username);
	if (r) {
		log_msg(LOG_ERR, "Error setting username");
		return NULL;
	}
	r = dbi_conn_set_option(conn, "password", password);
	if (r) {
		log_msg(LOG_ERR, "Error setting password");
		return NULL;
	}
	r = dbi_conn_set_option(conn, "dbname", dbname);
	if (r) {
		log_msg(LOG_ERR, "Error setting dbname");
		return NULL;
	}
	r = dbi_conn_set_option_numeric(conn, "port", port);
	if (r) {
		log_msg(LOG_ERR, "Error setting dbname");
		return NULL;
	}
	if (dbi_conn_connect(conn) < 0) {
		log_msg(LOG_ERR, "Could not connect. Please check the option settings.");
		return NULL;
	}
	connection = conn;
	return conn;
}

int
db_disconnect()
{
	dbi_conn_close(connection);
	dbi_shutdown();
	connection = NULL;
	return 0;
}

proc_handler *
get_handlers(const char *oid)
{
	proc_handler *handler = NULL, *prev = NULL, *first = NULL;
	char *o;
	dbi_conn_quote_string_copy(connection, oid, &o);
	dbi_result dbr = dbi_conn_queryf(connection,
			"SELECT name, script, context FROM handlers "
			"INNER JOIN matches ON handlers.id = matches.handler_id "
			"WHERE %s LIKE REPLACE(matches.oid, '*', '%%') "
			"ORDER BY matches.n, oid",
			o);
	free(o);
	if (!dbr || !dbi_result_get_numrows(dbr)) {
		dbi_result_free(dbr);
		dbr = dbi_conn_query(connection,
				"SELECT name, script, context FROM handlers "
				"INNER JOIN matches ON handlers.id = matches.handler_id "
				"WHERE matches.oid = \"fallback\" "
				"ORDER BY matches.n");
	}
	if (!dbr)
		return NULL;
	while (dbi_result_next_row(dbr)) {
		prev = handler;
		handler = calloc(1, sizeof(proc_handler));
		handler->name = dbi_result_get_string_copy(dbr, "name");
		handler->script = dbi_result_get_string_copy(dbr, "script");
		handler->context = dbi_result_get_string_copy(dbr, "context");
		if (prev)
			prev->next = handler;
		else
			first = handler;
	}
	dbi_result_free(dbr);
	return first;
}

char *
escape(const char *in)
{
	char *ret;
	dbi_conn_quote_string_copy(connection, in, &ret);
	return ret;
}

int
save_context(const char *context, const int handler_id)
{
	char *ctx;
	dbi_conn_quote_string_copy(connection, context, &ctx);
	dbi_result dbr = dbi_conn_queryf(connection,
			"UPDATE handlers "
			"SET context = %s "
			"WHERE id = %d",
			ctx,
			handler_id);
	free(ctx);
	dbi_result_free(dbr);
	return 0;
}

char *
get_module(const char *modulename)
{
	char *mname;
	dbi_conn_quote_string_copy(connection, modulename, &mname);
	dbi_result dbr = dbi_conn_queryf(connection,
			"SELECT script "
			"FROM modules "
			"WHERE name = %s "
			"LIMIT 1",
			mname);
	if (!dbr) {
		free(mname);
		return NULL;
	}
	if (!dbi_result_get_numrows(dbr)) {
		dbi_result_free(dbr);
		free(mname);
		return NULL;
	}
	dbi_result_next_row(dbr);
	char *script = dbi_result_get_string_copy(dbr, "script");
	dbi_result_free(dbr);
	free(mname);
	return script;
}
